@@ -467,6 +467,25 @@ def wp_draw_callback(points):
     mpstate.status.loading_waypoint_lasttime = time.time()
     mpstate.master().waypoint_count_send(mpstate.status.wploader.count())
 
+def wp_loop():
+    '''close the loop on a mission'''
+    loader = mpstate.status.wploader
+    if loader.count() < 2:
+        print("Not enough waypoints (%u)" % loader.count())
+        return
+    wp = loader.wp(loader.count()-2)
+    if wp.command == mavutil.mavlink.MAV_CMD_DO_JUMP:
+        print("Mission is already looped")
+        return
+    wp = mavutil.mavlink.MAVLink_mission_item_message(0, 0, 0, 0, mavutil.mavlink.MAV_CMD_DO_JUMP,
+                                                      0, 1, 1, -1, 0, 0, 0, 0, 0)
+    loader.add(wp)
+    loader.add(loader.wp(1))
+    mpstate.status.loading_waypoints = True
+    mpstate.status.loading_waypoint_lasttime = time.time()
+    mpstate.master().waypoint_count_send(mpstate.status.wploader.count())
+    print("Closed loop on mission")
+
 def set_home_location():
     '''set home location from last map click'''
     try:
@@ -492,7 +511,7 @@ def set_home_location():
 def cmd_wp(args):
     '''waypoint commands'''
     if len(args) < 1:
-        print("usage: wp <list|load|update|save|set|clear>")
+        print("usage: wp <list|load|update|save|set|clear|loop>")
         return
 
     if args[0] == "load":
@@ -542,6 +561,8 @@ def cmd_wp(args):
         print("Drawing waypoints on map")
     elif args[0] == "sethome":
         set_home_location()        
+    elif args[0] == "loop":
+        wp_loop()        
     else:
         print("Usage: wp <list|load|save|set|show|clear>")
 
@@ -736,7 +757,7 @@ def list_rally_points():
 
     for i in range(mpstate.status.rallyloader.rally_count()):
         p = mpstate.status.rallyloader.rally_point(i)
-        mpstate.console.writeln("lat=%f lng=%f alt=%f break_alt=%f land_dir=%f" % (p.lat * 1e-7, p.lng * 1e-7, p.alt / 100.0, p.break_alt / 100.0, p.land_dir / 100.0))
+        mpstate.console.writeln("lat=%f lng=%f alt=%f break_alt=%f land_dir=%f" % (p.lat * 1e-7, p.lng * 1e-7, p.alt, p.break_alt, p.land_dir))
 
     #TODO: rally_land points
 
@@ -765,19 +786,19 @@ def cmd_rally(args):
             print("No map click position available")
             return
 
-        alt = float(args[1]) * 100.0;
+        alt = float(args[1]);
         break_alt = 0.0
         land_hdg = 0.0;
         if (len(args) > 2):
-            break_alt = float(args[2]) * 100.0;
+            break_alt = float(args[2]);
         if (len(args) > 3):
-            land_hdg = float(args[3]) * 100.0;
+            land_hdg = float(args[3]);
 
         mpstate.status.rallyloader.create_and_append_rally_point(latlon[0] * 1e7, latlon[1] * 1e7, alt, break_alt, land_hdg, 0)
 
         send_rally_points();
 
-        print("Added Rally point at %s %f" % (str(latlon), alt / 100.0))
+        print("Added Rally point at %s %f" % (str(latlon), alt))
     
     elif args[0] == "clear":
         mpstate.status.rallyloader.clear()
@@ -1055,6 +1076,14 @@ def cmd_arm(args):
   '''arm motors'''
   mpstate.master().arducopter_arm()
 
+def cmd_time(args):
+  '''show autopilot time'''
+  tusec = mpstate.master().field('SYSTEM_TIME', 'time_unix_usec', 0)
+  if tusec == 0:
+      print("No SYSTEM_TIME time available")
+      return
+  print("%s (%s)\n" % (time.ctime(tusec * 1.0e-6), time.ctime()))
+
 
 def cmd_disarm(args):
   '''disarm motors'''
@@ -1105,6 +1134,7 @@ command_map = {
     'module'  : (cmd_module,   'module commands'),
     'alias'   : (cmd_alias,    'command aliases'),
     'arm'     : (cmd_arm,      'ArduCopter arm motors'),
+    'time'    : (cmd_time,     'Show autopilot time'),
     'disarm'  : (cmd_disarm,   'ArduCopter disarm motors')
     }
 
@@ -1729,6 +1759,15 @@ def check_link_status():
             say("link %u down" % (master.linknum+1))
             master.linkerror = True
 
+def send_heartbeat(master):
+    if master.mavlink10():
+        master.mav.heartbeat_send(mavutil.mavlink.MAV_TYPE_GCS, mavutil.mavlink.MAV_AUTOPILOT_INVALID,
+                                  0, 0, 0)
+    else:
+        MAV_GROUND = 5
+        MAV_AUTOPILOT_NONE = 4
+        master.mav.heartbeat_send(MAV_GROUND, MAV_AUTOPILOT_NONE)
+
 def periodic_tasks():
     '''run periodic checks'''
     if mpstate.status.setup_mode:
@@ -1740,13 +1779,7 @@ def periodic_tasks():
     if heartbeat_period.trigger() and mpstate.settings.heartbeat != 0:
         mpstate.status.counters['MasterOut'] += 1
         for master in mpstate.mav_master:
-            if master.mavlink10():
-                master.mav.heartbeat_send(mavutil.mavlink.MAV_TYPE_GCS, mavutil.mavlink.MAV_AUTOPILOT_INVALID,
-                                          0, 0, 0)
-            else:
-                MAV_GROUND = 5
-                MAV_AUTOPILOT_NONE = 4
-                master.mav.heartbeat_send(MAV_GROUND, MAV_AUTOPILOT_NONE)
+            send_heartbeat(master)
 
     if heartbeat_check_period.trigger():
         check_link_status()
@@ -1803,6 +1836,7 @@ def main_loop():
     '''main processing loop'''
     if not mpstate.status.setup_mode and not opts.nowait:
         for master in mpstate.mav_master:
+            send_heartbeat(master)
             master.wait_heartbeat()
             if len(mpstate.mav_param) < 10 or not mpstate.continue_mode:
                 mpstate.mav_param_set = set()
